// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// EncoderLame.cpp
//
// $Id: EncoderLame.cpp,v 1.6 2002/03/11 23:35:06 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "EncoderLame.h"

#include <string>
#include <xb/ErrorManager.h>
#include <xb/FileUtil.h>
#include <xb/String.h>

#include "portableio.h"
#include "PropList.h"
#include "System.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

const unsigned int EncoderLame::MAX_U_32_NUM;
const int EncoderLame::WAV_ID_WAVE;
const int EncoderLame::WAV_ID_FMT;
const int EncoderLame::WAV_ID_DATA;

//------------------------------------------------------------------------------

EncoderLame::EncoderLame()
{
	EM::I()->Noise() << "lame version:  " << get_lame_version() << endl;

	Init();
}

//------------------------------------------------------------------------------

EncoderLame::~EncoderLame()
{

}

//------------------------------------------------------------------------------

std::string
EncoderLame::EncodeWAV(const std::string& strFilename)
{
	m_strInFilename = strFilename;
	InitFiles();
	if (!m_pInFile || !m_pOutFile) {
		EM::I()->Error() << "file init failed, aborting encoding." << endl;
		throw(xb::file_error("file init failed, aborting encoding."));
	}

	int nResult = lame_init_params(m_pLameFlags);

	if (nResult < 0) {
		EM::I()->Error() << "error in lame init" << endl;
		throw(runtime_error("error in lame init"));
	}

	LameEncode();
	
	fclose(m_pOutFile);
	lame_close(m_pLameFlags);

	Init();
	
	return m_strOutFilename;
}

//------------------------------------------------------------------------------

void
EncoderLame::Init()
{
	m_pLameFlags = lame_init();
	AssertNotNull(m_pLameFlags);

	int nResult = lame_init_params(m_pLameFlags);
	Assert(nResult >= 0, "lame_init_params failure");

	m_nPCMBitWidth = 16;
	m_bCountSamplesCarefully = true;
	m_nNumSamplesRead = 0;
}

//------------------------------------------------------------------------------

void
EncoderLame::InitFiles()
{
	InitInFile();
	InitOutFile();
}

//------------------------------------------------------------------------------

void
EncoderLame::InitInFile()
{
	m_pInFile = OpenSoundFile(m_strInFilename);
}

//------------------------------------------------------------------------------

void
EncoderLame::InitOutFile()
{
	xb::FileUtil::AssertDirectoryExists(System::GetPropList()->GetValue("SpoolRoot"));
	m_strOutFilename = System::GetPropList()->GetValue("SpoolRoot") + "/tmp.mp3";

	m_pOutFile = fopen(m_strOutFilename.c_str(), "wb+");
	if (m_pOutFile == NULL) {
		throw(xb::file_error(std::string("couldn't open ") + m_strOutFilename +
						 " for writing..."));
	}
}

//------------------------------------------------------------------------------

FILE*
EncoderLame::OpenSoundFile(const std::string& strFilename)
{
	FILE* pResult = NULL;
	
	lame_set_num_samples(m_pLameFlags, MAX_U_32_NUM);
	pResult = fopen(m_strInFilename.c_str(), "rb");
	if (pResult == NULL) {
		EM::I()->Error() << "Couldn't open " << strFilename << endl;
		throw(xb::file_error(std::string("Couldn't open ") + strFilename));
	}

	ParseFileHeader(pResult);

	if (lame_get_num_samples(m_pLameFlags) == MAX_U_32_NUM) {

		double flen = xb::FileUtil::GetSize(strFilename);
		if (flen >= 0) {
			double fSamples = flen / (2 * lame_get_num_channels(m_pLameFlags));
			lame_set_num_samples(m_pLameFlags, (unsigned long)fSamples);
		}
	}

	return pResult;
}

//------------------------------------------------------------------------------

bool
EncoderLame::ParseFileHeader(FILE* pFile)
{
    int     nFormatTag = 0;
    int     nChannels = 0;
    int     nBlockAlign = 0;
    int     nBitsPerSample = 0;
    int     nSamplesPerSec = 0;
    int     nAvgBytesPerSec = 0;


    bool    bIsWav = false;
    long    nlDataLength = 0;
	long    nlSubSize = 0;

    // long nlFileLength = Read32BitsHighLow(pFile);

    if (Read32BitsHighLow(pFile) != WAV_ID_WAVE) {
        return false;
	}

    for (int nLoopSanity = 0; nLoopSanity < 20; ++nLoopSanity) {
        int nType = Read32BitsHighLow(pFile);

        if (nType == WAV_ID_FMT) {
            nlSubSize = Read32BitsLowHigh(pFile);
            if (nlSubSize < 16) {
                /*DEBUGF(
                   "'fmt' chunk too short (only %ld bytes)!", nlSubSize);  */
                return false;
            }

            nFormatTag = Read16BitsLowHigh(pFile);
            nlSubSize -= 2;
            nChannels = Read16BitsLowHigh(pFile);
            nlSubSize -= 2;
            nSamplesPerSec = Read32BitsLowHigh(pFile);
            nlSubSize -= 4;
            nAvgBytesPerSec = Read32BitsLowHigh(pFile);
            nlSubSize -= 4;
            nBlockAlign = Read16BitsLowHigh(pFile);
            nlSubSize -= 2;
            nBitsPerSample = Read16BitsLowHigh(pFile);
            nlSubSize -= 2;

            /* DEBUGF("   skipping %d bytes\n", nlSubSize); */

            if (nlSubSize > 0) {
                if (!xb::FileUtil::Skip(pFile, (long) nlSubSize, SEEK_CUR)) {
                    return false;
				}
            };

        }
        else if (nType == WAV_ID_DATA) {
            nlSubSize = Read32BitsLowHigh(pFile);
            nlDataLength = nlSubSize;
			bIsWav = true;
            /* We've found the audio data. Read no further! */
            break;

        }
        else {
            nlSubSize = Read32BitsLowHigh(pFile);
            if (!xb::FileUtil::Skip(pFile, (long) nlSubSize, SEEK_CUR)) {
                return 0;
			}
        }
    }

    if (nFormatTag != 1) {
		return false; /* oh no! non-supported format  */
    }


    if (bIsWav) {
        /* make sure the header is sane */
        if (-1 == lame_set_num_channels(m_pLameFlags, nChannels)) {
			throw(runtime_error(std::string("Unsupported number of channels: ") +
								xb::String(nChannels)));
        }
        (void) lame_set_in_samplerate( m_pLameFlags, nSamplesPerSec );
        m_nPCMBitWidth = nBitsPerSample;
        (void) lame_set_num_samples( m_pLameFlags,
									 nlDataLength /
									 (nChannels * ((nBitsPerSample + 7) / 8)) );
    }
	
    return bIsWav;
}

//------------------------------------------------------------------------------

void
EncoderLame::LameEncode()
{
	// int silent = 0;
	
    unsigned char mp3buffer[LAME_MAXMP3BUFFER];
    int     Buffer[2][1152];
    int     iread, imp3;
    int     frames;

	
    /* encode until we hit eof */
    do {
        /* read in 'iread' samples */
        iread = GetAudio(Buffer);
        frames = lame_get_frameNum(m_pLameFlags);

        /* encode */
        imp3 = lame_encode_buffer_int(m_pLameFlags, Buffer[0], Buffer[1], iread,
                                      mp3buffer, sizeof(mp3buffer));

        /* was our output buffer big enough? */
        if (imp3 < 0) {
            if (imp3 == -1) {
                throw(runtime_error("mp3 buffer is not big enough..."));
			}
            else {
                throw(runtime_error("mp3 internal error:  error code=" +
									xb::String(imp3)));
			}
        }

        if ((int)fwrite(mp3buffer, 1, imp3, m_pOutFile) != imp3) {
			throw(xb::file_error("Error writing mp3 output"));
        }

    } while (iread);

	imp3 = lame_encode_flush(m_pLameFlags, mp3buffer, sizeof(mp3buffer)); /* may return one more mp3 frame */

    if (imp3 < 0) {
        if (imp3 == -1) {
			throw(runtime_error("mp3 buffer is not big enough..."));
		}
        else {
            throw(runtime_error("mp3 internal error:  error code=" +
								xb::String(imp3)));
		}
    }

    fwrite(mp3buffer, 1, imp3, m_pOutFile);
}

//------------------------------------------------------------------------------

int
EncoderLame::GetAudio(int anBuffer[2][1152])
{
    int     num_channels = lame_get_num_channels( m_pLameFlags );
    int     insamp[2 * 1152];
    int     samples_read;
    int     framesize;
    int     samples_to_read;
    unsigned int remaining, tmp_num_samples;
    int     i;
    int     *p;

    /* 
     * NOTE: LAME can now handle arbritray size input data packets,
     * so there is no reason to read the input data in chuncks of
     * size "framesize".  EXCEPT:  the LAME graphical frame analyzer 
     * will get out of sync if we read more than framesize worth of data.
     */

    samples_to_read = framesize = lame_get_framesize(m_pLameFlags);
    Assert(framesize <= 1152, "framesize too large");

    /* get num_samples */
    tmp_num_samples = lame_get_num_samples( m_pLameFlags );

    /* if this flag has been set, then we are carefull to read
     * exactly num_samples and no more.  This is useful for .wav and .aiff
     * files which have id3 or other tags at the end.  Note that if you
     * are using LIBSNDFILE, this is not necessary 
     */
    if (m_bCountSamplesCarefully) {
        remaining = tmp_num_samples - std::min(tmp_num_samples, m_nNumSamplesRead);
        if ((int)remaining < framesize)
            samples_to_read = remaining;
    }

	samples_read = ReadSamplesPCM(insamp, num_channels * framesize,
								  num_channels * samples_to_read);
	p = insamp + samples_read;
	samples_read /= num_channels;
	if( anBuffer != NULL ) {	/* output to int buffer */
	    if( num_channels == 2 ) {
			for( i = samples_read; --i >= 0; ) {
				anBuffer[1][i] = *--p;
				anBuffer[0][i] = *--p;
			}
	    } else if( num_channels == 1 ) {
			memset( anBuffer[1], 0, samples_read * sizeof(int) );
			for( i = samples_read; --i >= 0; ) {
				anBuffer[0][i] = *--p;
			}
	    } else {
			Assert(0, "invalid num_channels");
		}
	}

    /* if num_samples = MAX_U_32_NUM, then it is considered infinitely long.
       Don't count the samples */
    if ( tmp_num_samples != MAX_U_32_NUM )
        m_nNumSamplesRead += samples_read;

    return samples_read;
}

//------------------------------------------------------------------------------

int
EncoderLame::ReadSamplesPCM(int anSampleBuffer[2304], int nFrameSize,
						int nSamplesToRead)
{
    int     samples_read;
    int     iswav = 1;
	int     swapbytes = 0;
    int     hi_lo_order;	/* byte order of input stream */

    if( (24 == m_nPCMBitWidth) || (16 == m_nPCMBitWidth) ) {
				/* assume only recognized wav files are */
				/*  in little endian byte order */
		hi_lo_order = (!iswav == !swapbytes);
		if( 16 == m_nPCMBitWidth )
			samples_read = UnpackReadSamples(nSamplesToRead, 2, hi_lo_order,
											 anSampleBuffer, m_pInFile );
		else			/* ( 24 == m_nPCMBitWidth ) */
			samples_read = UnpackReadSamples(nSamplesToRead, 3, hi_lo_order,
											 anSampleBuffer, m_pInFile );
    }
	else if (8 == m_nPCMBitWidth) {
		samples_read = UnpackReadSamples(nSamplesToRead, 1, 0,
										 anSampleBuffer, m_pInFile );
    }
	else {
		EM::I()->Error() << "Only 8, 16, and 24 bit input files supported" <<
			endl;
		throw(invalid_argument("Only 8, 16, and 24 bit input files supported"));
    }
	
    if (ferror(m_pInFile)) {
		EM::I()->Error() << "Error reading input file" << endl;
		throw(xb::file_error("Error reading input file"));
    }

    return samples_read;
}

//------------------------------------------------------------------------------

int
EncoderLame::UnpackReadSamples(int nSamplesToRead, int nBytesPerSample,
						   int nSwapOrder, int* pSampleBuffer,
						   FILE* pPCMin )
{
    int samples_read;
    int i;
    int *op;			/* output pointer */
    unsigned char *ip = (unsigned char *) pSampleBuffer; /* input pointer */
    const int b = sizeof(int) * 8;

#define GA_URS_IFLOOP( ga_urs_bps ) \
    if( nBytesPerSample == ga_urs_bps ) \
	for( i = samples_read * nBytesPerSample; (i -= nBytesPerSample) >=0;)
    

    samples_read = fread( pSampleBuffer, nBytesPerSample, 
			  nSamplesToRead, pPCMin);
    op = pSampleBuffer + samples_read;

    GA_URS_IFLOOP( 1 )
	*--op = (ip[i] ^ 0x80)<<(b-8) | 0x7f<<(b-16);/* convert from unsigned*/
    if( nSwapOrder == 0 ) {
	GA_URS_IFLOOP( 2 )
	    *--op = ip[i]<<(b-16) | ip[i+1]<<(b-8); 
	GA_URS_IFLOOP( 3 )
	    *--op = ip[i]<<(b-24) | ip[i+1]<<(b-16) | ip[i+2]<<(b-8);
    } else {
	GA_URS_IFLOOP( 2 )
	    *--op = ip[i]<<(b-8) | ip[i+1]<<(b-16); 
	GA_URS_IFLOOP( 3 )
	    *--op = ip[i]<<(b-8) | ip[i+1]<<(b-16) | ip[i+2]<<(b-24);
    }
#undef GA_URS_IFLOOP
    return( samples_read );
}

//------------------------------------------------------------------------------

extern "C" {
	Encoder* maker()
	{
		return new EncoderLame;
	}
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
