// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// Frame.cpp
//
// $Id: Frame.cpp,v 1.8 2002/02/26 19:30:26 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "Frame.h"

#include <xb/ErrorManager.h>

#include "Synth.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

const int Frame::s_TabSel[2][3][16] =
{
	{
    {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448,},
       {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384,},
       {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320,}},

	{
       {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256,},
	    {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160,},
	    {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160,}}
};

const long Frame::s_Freqs[9] =
{44100, 48000, 32000, 22050, 24000, 16000, 11025, 12000, 8000};

const int Frame::MAXFRAMESIZE;
const int Frame::MPG_MD_STEREO;
const int Frame::MPG_MD_JOINT_STEREO;
const int Frame::MPG_MD_DUAL_CHANNEL;
const int Frame::MPG_MD_MONO;

//------------------------------------------------------------------------------

Frame::Frame() :
	synth(NULL),
	synth_mono(NULL)
{
	Reset();
}

//------------------------------------------------------------------------------

Frame::~Frame()
{
	// delete synth;
	// delete synth_mono;

	// Reset();
}

//------------------------------------------------------------------------------

void
Frame::Reset()
{
	alloc = 0;
	synth = 0;
	synth_mono = 0;
		
#ifdef USE_3DNOW
	dct36 = 0;
#endif
	stereo = 0;
	jsbound = 0;
	single = 0;
	II_sblimit = 0;
	down_sample_sblimit = 0;
	lsf = 0;
	mpeg25 = 0;
	down_sample = 0;
	header_change = 0;
	lay = 0;
	do_layer = 0;
	error_protection = 0;
	bitrate_index = 0;
	sampling_frequency = 0;
	padding = 0;
	extension = 0;
	mode = 0;
	mode_ext = 0;
	copyright = 0;
	original = 0;
	emphasis = 0;
	framesize = 0;		/* computed framesize */
}

//------------------------------------------------------------------------------

double
Frame::ComputeTPF()
{
	static int bs[4] = {0, 384, 1152, 1152};
	double tpf;

	tpf = (double) bs[lay];
	tpf /= s_Freqs[sampling_frequency] << lsf;
	
	return tpf;
}

//------------------------------------------------------------------------------

double
Frame::ComputeBPF()
{
	double bpf;

	switch (lay)
	{
		case 1:
			bpf = s_TabSel[lsf][0][bitrate_index];
			bpf *= 12000.0 * 4.0;
			bpf /= s_Freqs[sampling_frequency] << (lsf);
			break;
		case 2:
		case 3:
			bpf = s_TabSel[lsf][lay - 1][bitrate_index];
			bpf *= 144000;
			bpf /= s_Freqs[sampling_frequency] << (lsf);
			break;
		default:
			bpf = 1.0;
	}

	return bpf;
}

//------------------------------------------------------------------------------

int
Frame::DecodeHeader(Frame* fr, unsigned long newhead)
{
	int nResult = 0;
	
	if (newhead & (1 << 20))
	{
		fr->lsf = (newhead & (1 << 19)) ? 0x0 : 0x1;
		fr->mpeg25 = 0;
	}
	else
	{
		fr->lsf = 1;
		fr->mpeg25 = 1;
	}
	fr->lay = 4 - ((newhead >> 17) & 3);
	if (fr->mpeg25)
	{
		fr->sampling_frequency = 6 + ((newhead >> 10) & 0x3);
	}
	else
		fr->sampling_frequency = ((newhead >> 10) & 0x3) + (fr->lsf * 3);
	fr->error_protection = ((newhead >> 16) & 0x1) ^ 0x1;

	if (fr->mpeg25)		/* allow Bitrate change for 2.5 ... */
		fr->bitrate_index = ((newhead >> 12) & 0xf);

	fr->bitrate_index = ((newhead >> 12) & 0xf);
	fr->padding = ((newhead >> 9) & 0x1);
	fr->extension = ((newhead >> 8) & 0x1);
	fr->mode = ((newhead >> 6) & 0x3);
	fr->mode_ext = ((newhead >> 4) & 0x3);
	fr->copyright = ((newhead >> 3) & 0x1);
	fr->original = ((newhead >> 2) & 0x1);
	fr->emphasis = newhead & 0x3;

	fr->stereo = (fr->mode == MPG_MD_MONO) ? 1 : 2;

	if (!fr->bitrate_index)
		return (0);

	switch (fr->lay)
	{
		case 1:
			// PATRICK: fix
			// fr->do_layer = PlayerMPG::DoLayer1;
			// InitLayer2();	/* inits also shared tables with layer1 */
			fr->framesize = (long) s_TabSel[fr->lsf][0][fr->bitrate_index] * 12000;
			fr->framesize /= s_Freqs[fr->sampling_frequency];
			fr->framesize = ((fr->framesize + fr->padding) << 2) - 4;
			break;
		case 2:
			// PATRICK: fix
			// fr->do_layer = PlayerMPG::DoLayer2;
			// InitLayer2();	/* inits also shared tables with layer1 */
			fr->framesize = (long) s_TabSel[fr->lsf][1][fr->bitrate_index] * 144000;
			fr->framesize /= s_Freqs[fr->sampling_frequency];
			fr->framesize += fr->padding - 4;
			break;
		case 3:
			// fr->do_layer = PlayerMPG::DoLayer3;
			if (fr->lsf)
				nResult = (fr->stereo == 1) ? 9 : 17;
			else
				nResult = (fr->stereo == 1) ? 17 : 32;
			if (fr->error_protection)
				nResult += 2;
			fr->framesize = (long) s_TabSel[fr->lsf][2][fr->bitrate_index] * 144000;
			fr->framesize /= s_Freqs[fr->sampling_frequency] << (fr->lsf);
			fr->framesize = fr->framesize + fr->padding - 4;
			break;
		default:
			return (0);
	}
	
	if(fr->framesize > MAXFRAMESIZE) {
		EM::I()->Warn() << "framesize too big" << endl;
		return 0;
	}
	
	return nResult;
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------

