// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// PlayerOgg.cpp
//
// $Id: PlayerOgg.cpp,v 1.6 2002/02/26 19:30:27 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "PlayerOgg.h"

#include <ogg/ogg.h>
#include <stdio.h>
#include <vorbis/codec.h>
#include <xb/ErrorManager.h>

#include "Output.h"
#include "Visualization.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

const ov_callbacks PlayerOgg::s_VorbisCallbacks = {
	/*
	fread,
	(int (*)(void *, long long int, int))fseek,
	fclose,
	ftell
	*/
	PlayerOgg::ovcb_read,
	PlayerOgg::ovcb_seek,
	PlayerOgg::ovcb_close,
	PlayerOgg::ovcb_tell
};

//------------------------------------------------------------------------------

PlayerOgg::PlayerOgg() :
	m_bEOS(false),
	m_nBytesStreamed(0),
	m_bOutputError(false),
	m_nTrack(0),
	m_nSampleRate(0),
	m_nChannels(0),
	m_nLength(0)
{

}

//------------------------------------------------------------------------------

PlayerOgg::~PlayerOgg()
{

}

//------------------------------------------------------------------------------


void
PlayerOgg::Init(Output* pOutput, Visualization* pVis)
{
	Player::Init(pOutput, pVis);
}

//------------------------------------------------------------------------------

void
PlayerOgg::Play(const xb::Filename& strFilename, int nTrack)
{
	m_strFilename = strFilename;
	m_nTrack = nTrack;

	m_bPlaying = true;
	m_bEOS = false;
	m_nBytesStreamed = 0;
	m_bOutputError = false;
	
	Start();
}

//------------------------------------------------------------------------------

void
PlayerOgg::Stop(bool bSendSignal = true)
{
	if (m_bPlaying) {
		Terminate();
		m_pOutput->Close();
		m_bPlaying = false;
	}
}

//------------------------------------------------------------------------------

void
PlayerOgg::Pause(bool bPause)
{
	m_pOutput->Pause(bPause);
}

//------------------------------------------------------------------------------

void
PlayerOgg::Seek(int nTime)
{
	m_nSeekNeeded = nTime;

	while (m_nSeekNeeded != -1) {
		Thread::usleep(20000);
	}
}

//------------------------------------------------------------------------------

void
PlayerOgg::SeekPercentage(float fPercent)
{
	int nTime = (int)(m_nLength * fPercent / 1000.0);
	Seek(nTime);
}

//------------------------------------------------------------------------------

int
PlayerOgg::GetTime() const
{
	if (m_bOutputError) {
		return -2;
	}

	if (!m_bPlaying || (m_bEOS && !m_pOutput->BufferPlaying())) {
		return -1;
	}

	return m_pOutput->ElapsedTime();
}

//------------------------------------------------------------------------------

bool
PlayerOgg::CanPlayFile(const xb::Filename& strFilename)
{
	bool bResult = false;

	FILE* pStream = fopen(strFilename.c_str(), "rb");
	if (pStream != NULL) {

		OggVorbis_File vfile;
		memset(&vfile, 0, sizeof(vfile));

		ost::MutexLock lock(m_Mutex);
		if (ov_open(pStream, &vfile, NULL, 0) == 0) {
			EM::I()->Trace() << "PlayerOgg::CanPlayFile() = true" << endl;
			bResult = true;
			ov_clear(&vfile);
		}
		else {
			EM::I()->Trace() << "PlayerOgg::CanPlayFile() = false" << endl;
			fclose(pStream);
		}

	}
	
	return bResult;
}

//------------------------------------------------------------------------------

bool
PlayerOgg::CheckCapability(std::string& strMessage)
{
	return true;
}

//------------------------------------------------------------------------------

void
PlayerOgg::Run()
{
	EM::I()->Trace() << "PlayerOgg::Run()" << endl;

	EM::I()->Trace() << "PlayerOgg::Run() -- opening file:  " <<
		m_strFilename << endl;
	
	if (m_strFilename.length() <= 0) {
		return;
	}
	
	FILE* pStream = fopen(m_strFilename.c_str(), "rb");
	if (pStream == NULL) {
		EM::I()->Trace() << "PlayerOgg::Run() -- file error." << endl;
		m_bEOS = true;
		// PATRICK: fix
		// goto play_cleanup
		return;
	}

	EM::I()->Trace() << "PlayerOgg::Run() -- file ok." << endl;

	// block for lock
	{
		ost::MutexLock lock(m_Mutex);

		// if (ov_open_callbacks(pStream, &m_VF, NULL, 0, s_VorbisCallbacks) < 0) {
		EM::I()->Trace() << "PlayerOgg::Run() -- ov_open..." << endl;
		if (ov_open(pStream, &m_VF, NULL, 0) < 0) {
			//OggVorbis_File vf;
			//if (ov_open(pStream, &vf, NULL, 0) < 0) {
			EM::I()->Trace() << "PlayerOgg::Run() -- ov_open failed." << endl;
			fclose(pStream);
			m_bEOS = true;
			// PATRICK: fix
			// goto play_cleanup
			return;
		}

		EM::I()->Trace() << "PlayerOgg::Run() -- ov_open ok." << endl;

		m_nLength = (int)ov_time_total(&m_VF, -1) * 1000;
		
		vorbis_info* pVI = ov_info(&m_VF, -1);
	
		if (pVI->channels > 2) {
			fclose(pStream);
			m_bEOS = true;
			// PATRICK: fix
			// goto play_cleanup
			return;
		}

		// double dTime = ov_time_total(&m_VF, -1) * 1000;
	
		m_nSampleRate = pVI->rate;
		m_nChannels = pVI->channels;

	// title = vorbis_generate_title(&vf, filename);
	// vorbis_ip.set_info(title, time, ov_bitrate(&vf, -1), samplerate, channels);
		if (!m_pOutput->Open(Output::FMT_S16_NE, m_nSampleRate, m_nChannels,
							 m_nTrack))
		{
			fclose(pStream);
			m_bOutputError = true;
			m_bEOS = true;
			// PATRICK: fix
			// goto play_cleanup
			return;
		
		}
	} // lock released

	m_nSeekNeeded = -1;

	int nLastSection = -1;
	
	while (m_bPlaying) {
		if (m_nSeekNeeded != -1) {
			DoSeek();
		}

		if (m_bEOS) {
			Thread::usleep(20000);
			continue;
		}

		int nCurrentSection = ProcessData(nLastSection);

		if (nCurrentSection != nLastSection) {
			/*
			 * set total play time, bitrate, rate, and channels of
			 * current section

			if (title)
				g_free(title);
			pthread_mutex_lock(&vf_mutex);
			title = vorbis_generate_title(&vf, filename);
			
			if (vorbis_is_streaming)
				time = -1;
			else
				time = ov_time_total(&vf, -1) * 1000;

			vorbis_ip.set_info(title, time,
					   ov_bitrate(&vf, current_section), 
					   samplerate, channels);
			pthread_mutex_unlock(&vf_mutex);
			*/

			// timercount = m_pOutput->ElapsedTime();

			nLastSection = nCurrentSection;
		}

		/*
		if (!(vi->bitrate_upper == vi->bitrate_lower == vi->bitrate_nominal) &&
		     (vorbis_ip.output->output_time() > timercount + 1000 ||
		      vorbis_ip.output->output_time() < timercount)) {
			
			  // simple hack to avoid updating too
			  // often

			long br;
			
			pthread_mutex_lock(&vf_mutex);
			br = ov_bitrate_instant(&vf);
			pthread_mutex_unlock(&vf_mutex);
			if (br > 0)
				vorbis_ip.set_info(title, time, br,
						   samplerate, channels);
			timercount = vorbis_ip.output->output_time();
		}
		*/

		
		
	}

	if (!m_bOutputError) {
		m_pOutput->Close();
	}

	{
		ost::MutexLock lock(m_Mutex);
		ov_clear(&m_VF);
	}
	
	EM::I()->Trace() << "PlayerOgg::Run() -- done" << endl;
}

//------------------------------------------------------------------------------

#ifdef WORDS_BIGENDIAN
#define host_is_big_endian() TRUE
#else
#define host_is_big_endian() FALSE
#endif

//------------------------------------------------------------------------------

int
PlayerOgg::ProcessData(int nLastSection)
{
	char pcmout[4096];
	int bep = host_is_big_endian();
	int nResult = 0;
	int nCurrentSection;
	
	{
		ost::MutexLock lock(m_Mutex);

		nResult = ov_read(&m_VF, pcmout, sizeof(pcmout), bep, 2, 1,
							  &nCurrentSection);
		
		switch (nResult) {
			case 0:
				// EOF
				m_pOutput->BufferFreeAmt();
				m_pOutput->BufferFreeAmt();
				m_bEOS = true;
				return nLastSection;
			case -1:
				// stream error, can continue
				return nLastSection;
		}

		if (nCurrentSection != nLastSection) {
			vorbis_info* pVI = ov_info(&m_VF, -1);

			if (pVI->channels > 2) {
				m_bEOS = true;
				return nCurrentSection;
			}

			if (pVI->rate != m_nSampleRate || pVI->channels != m_nChannels) {
				m_nSampleRate = pVI->rate;
				m_nChannels = pVI->channels;
				m_pOutput->BufferFreeAmt();
				m_pOutput->BufferFreeAmt();
				m_pOutput->Close();
				if (!m_pOutput->Open(Output::FMT_S16_NE, m_nSampleRate, m_nChannels,
									 m_nTrack))	{
					m_bOutputError = true;
					m_bEOS = true;
					return nCurrentSection;
				}

				m_pOutput->Flush(ov_time_tell(&m_VF) * 1000);
			}
		}
	}		// release lock
	
	m_pVisualization->SetData(m_pOutput->WrittenTime(),
							  Output::FMT_S16_NE, m_nChannels, nResult, pcmout);

	while (m_pOutput->BufferFreeAmt() < nResult) {
		Thread::usleep(20000);
		if (!m_bPlaying) {
			return nCurrentSection;
		}

		if (m_nSeekNeeded != -1) {
			DoSeek();
		}
	}

	m_pOutput->Write(pcmout, nResult);

	return nCurrentSection;

}

//------------------------------------------------------------------------------

void
PlayerOgg::DoSeek()
{
	if (m_nSeekNeeded != -1) {
		JumpToTime(m_nSeekNeeded);
		m_nSeekNeeded = -1;
		m_bEOS = false;
	}
}

//------------------------------------------------------------------------------

void
PlayerOgg::JumpToTime(long nTime)
{
	ost::MutexLock lock(m_Mutex);

	// PATRICK: >= ???
	if (nTime == ov_time_total(&m_VF, -1)) {
		nTime--;
	}

	m_pOutput->Flush(nTime * 1000);
	ov_time_seek(&m_VF, nTime);
}

//------------------------------------------------------------------------------

size_t
PlayerOgg::ovcb_read(void *ptr, size_t size, size_t nmemb, void *datasource)
{
	return fread(ptr, size, nmemb, (FILE *)datasource);
}

//-------------------------------------------------------------------------------

int
PlayerOgg::ovcb_seek(void *datasource, int64_t offset, int whence)
{
	return fseek((FILE *)datasource, offset, whence);
}

//------------------------------------------------------------------------------

int
PlayerOgg::ovcb_close(void *datasource)
{
	return fclose((FILE *)datasource);
}

//------------------------------------------------------------------------------

long
PlayerOgg::ovcb_tell(void *datasource)
{
	return ftell((FILE *)datasource);
}

//------------------------------------------------------------------------------

extern "C" {
	Player* maker()
	{
		return new PlayerOgg;
	}
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
