// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// RipperCD.cpp
//
// $Id: RipperCD.cpp,v 1.21 2002/02/27 07:35:49 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "RipperCD.h"

#include <stdio.h>
#include <xb/ErrorManager.h>
#include <xb/Filename.h>
#include <xb/FileUtil.h>
#include <xb/String.h>

#include "Output.h"
#include "PropList.h"
#include "System.h"
#include "Thread.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

const int RipperCD::OUTBUFSZ;

//------------------------------------------------------------------------------

RipperCD::RipperCD() :
	m_pDrive(NULL),
	m_pParanoia(NULL),
	m_pOutput(NULL)
{	
}

//------------------------------------------------------------------------------

RipperCD::~RipperCD()
{
	if (m_pParanoia != NULL) {
		paranoia_free(m_pParanoia);
	}
	
	if (m_pDrive != NULL) {
		cdda_close(m_pDrive);
	}
}

//------------------------------------------------------------------------------

bool
RipperCD::Init(const std::string& strDevice, Output* pOutput)
{
	bool bResult = true;

	m_pOutput = pOutput;
	AssertNotNull(m_pOutput);
	
	int nVerbose = CDDA_MESSAGE_PRINTIT;

	if (strDevice == "auto") {
		m_pDrive = cdda_find_a_cdrom(nVerbose, NULL);
		if (m_pDrive == NULL) {
			throw(runtime_error("couldn't find a cdrom drive"));
		}
	}
	else {
		m_pDrive = cdda_identify(strDevice.c_str(), nVerbose, NULL);
		if (m_pDrive == NULL) {
			throw(runtime_error("couldn't find a valid cdrom drive at " +
				strDevice));
		}
	}

	if (nVerbose) {
		cdda_verbose_set(m_pDrive, CDDA_MESSAGE_PRINTIT, CDDA_MESSAGE_PRINTIT);
	}
	else {
		cdda_verbose_set(m_pDrive, CDDA_MESSAGE_PRINTIT, CDDA_MESSAGE_FORGETIT);
	}

	int nSpeed = System::GetPropList()->GetIntValue("RipSpeed");
	int nResult = cdda_speed_set(m_pDrive, nSpeed);

	
	nResult = cdda_open(m_pDrive);
	switch(nResult) {
		case -2:
		case -3:
		case -4:
		case -5:
			EM::I()->Error() <<
				"Unable to open disc.  Is there an audio CD in the drive?" <<
				endl;
			bResult = false;
			return(bResult);
		case -6:
			EM::I()->Error() << "No way to read audio from this drive." << endl;
			bResult = false;
			return(bResult);
		case 0:
			break;
		default:
			EM::I()->Error() << "Unable to open disc." << endl;
			bResult = false;
			return(bResult);
	}

	if (nVerbose) {
		DisplayTOC();
	}
	
	return bResult;
}

//------------------------------------------------------------------------------

void
RipperCD::DisplayTOC()
{
	if (m_pDrive != NULL) {
		long audiolen = 0;
		cout << endl << "Table of contents (audio tracks only):" << endl <<
			"track        length               begin        copy pre ch" << endl << 
			"===========================================================" << endl;
  
		for (int i = 1; i <= m_pDrive->tracks; i++) {
			if (cdda_track_audiop(m_pDrive, i)) {
				char buffer[256];

				long sec = cdda_track_firstsector(m_pDrive, i);
				long off = cdda_track_lastsector(m_pDrive, i) - sec + 1;
      
				sprintf(buffer,
						"%3d.  %7ld [%02d:%02d.%02d]  %7ld [%02d:%02d.%02d]  %s %s %s",
						i,
						off,(int)(off/(60*75)),(int)((off/75)%60),(int)(off%75),
						sec,(int)(sec/(60*75)),(int)((sec/75)%60),(int)(sec%75),
						cdda_track_copyp(m_pDrive, i) ? "  OK" : "  no",
						cdda_track_preemp(m_pDrive, i) ? " yes" : "  no",
						cdda_track_channels(m_pDrive, i) == 2 ? " 2" : " 4");
				// report(buffer);
				cout << buffer << endl;
				audiolen += off;
			}
		}

		char buffer[256];
		sprintf(buffer, "TOTAL %7ld [%02d:%02d.%02d]    (audio only)",
				audiolen, (int)(audiolen / (60 * 75)), (int)((audiolen / 75) % 60),
				(int)(audiolen % 75));
		cout << buffer << endl;
	
		cout << endl;
	}
}

//------------------------------------------------------------------------------

std::string
RipperCD::Rip(int nTrack)
{
	AssertNotNull(m_pDrive);
	Assert(nTrack >= 0, "track < 0");
	Assert(nTrack <= m_pDrive->tracks, "track > num tracks");
	
	if (!cdda_track_audiop(m_pDrive, nTrack)) {
		throw(invalid_argument("Track " + xb::String(nTrack) + " is not an audio track."));
	}

	m_pParanoia = paranoia_init(m_pDrive);

	long nFirstSector = cdda_track_firstsector(m_pDrive, nTrack);
	long nLastSector = cdda_track_lastsector(m_pDrive, nTrack);

	long nCursor = nFirstSector;
	paranoia_seek(m_pParanoia, nCursor, SEEK_SET);

	seteuid(getuid());
	setegid(getgid());

	Assert(m_pOutput->IsFileOutput(),
		   "output plugin selected does not generate a file.");

	xb::FileUtil::AssertDirectoryExists(
		System::GetPropList()->GetValue("SpoolRoot"));
	
	xb::Filename strFilename = System::GetPropList()->GetValue("SpoolRoot") +
		"/track" + xb::String(nTrack) + m_pOutput->GetFileExtension();

	if (!m_pOutput->Open(Output::FMT_S16_NE, 44100, 2, nTrack)) {
		throw(xb::file_error("Cannot open output file"));
	}
	
	int skipped_flag = 0;
	int nMaxRetries = 20;
	int nAbortOnSkip = 1;
	int sample_offset = 0;
	int offset_skip = sample_offset * 4;
	
	while (nCursor <= nLastSector) {
		int16_t* readbuf = paranoia_read_limited(m_pParanoia, NULL, nMaxRetries);
		char* err = cdda_errors(m_pDrive);
		char* mes = cdda_messages(m_pDrive);

		if (mes || err) {
			fprintf(stderr, "\r                               "
					"                                           \r%s%s\n",
					mes ? mes : "", err ? err : "");
			
			if (err) { free(err); }
			if (mes) { free(mes); }
		}

		if (readbuf == NULL) {
			skipped_flag = 1;
			throw(xb::file_error("paranoia_read:  Unrecoverable error, bailing."));
		}

		if (skipped_flag && nAbortOnSkip) {
			nCursor = nLastSector + 1;
			break;
		}

		skipped_flag = 0;
		nCursor++;

		int nSize = CD_FRAMESIZE_RAW - offset_skip;

		while (m_pOutput->BufferFreeAmt() < nSize) {
			EM::I()->Trace() << "waiting for buffer to free up..." << endl;
			Thread::usleep(10000);
		}

		m_pOutput->Write((char *)readbuf + offset_skip, nSize);
		
		offset_skip = 0;
	}

	m_pOutput->Close();

	paranoia_free(m_pParanoia);
	m_pParanoia = NULL;

	EM::I()->Trace() << "Done." << endl;

	return strFilename;
}

//------------------------------------------------------------------------------

bool
RipperCD::CheckCapability(std::string& strMessage)
{
	bool bResult = true;
	int nVerbose = CDDA_MESSAGE_PRINTIT;
	
	cdrom_drive* pDrive = cdda_find_a_cdrom(nVerbose, NULL);

	if (pDrive == NULL) {
		bResult = false;
		strMessage = "Couldn't find a cdrom drive for ripping.";
	}
	else {
		cdda_close(pDrive);
	}
	
	return bResult;
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
