// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// SourceTable.cpp
//
// $Id: SourceTable.cpp,v 1.27 2002/03/12 00:04:51 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include "SourceTable.h"

#include <gtk--/style.h>
#include <list>
#include <xb/AutoLock.h>
#include <xb/AutoLockGtk.h>
#include <xb/ErrorManager.h>
#include <xb/Pixmap.h>

#include "Controller.h"
#include "InterfaceTypes.h"
#include "PlayListManager.h"
#include "System.h"

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

pthread_mutex_t SourceTable::s_ThreadLock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t SourceTable::s_AddRowLock = PTHREAD_MUTEX_INITIALIZER;

//------------------------------------------------------------------------------

SourceTable::SourceTable() :
	xb::XCList(2),
	m_nLastRow(0),
	m_nDragRow(-1),
	m_iconLibrary(INTERFACE_DATA_DIR "/icon-library.xpm"),
	m_iconCD(INTERFACE_DATA_DIR "/icon-cd.xpm"),
	m_iconPlayList(INTERFACE_DATA_DIR "/icon-list.xpm")
{
	set_shadow_type(GTK_SHADOW_ETCHED_IN);
	set_column_auto_resize(0, true);
	set_column_auto_resize(1, true);

	drag_dest_set(GTK_DEST_DEFAULT_ALL,
				  InterfaceTypes::s_aTargetTableStrings, 1,
				  (GdkDragAction)(GDK_ACTION_COPY));
	drag_data_received.connect(slot(this, &SourceTable::DNDReceive));

	// motion_notify_event.connect(slot(this, &SourceTable::Motion));
	drag_motion.connect(slot(this, &SourceTable::DragMotion));
	drag_leave.connect(slot(this, &SourceTable::DragLeave));

	System::GetController()->CDAvailableChange.connect(
		SigC::slot(this, &SourceTable::CDAvailableChange));

	Gdk_Pixmap pm = m_iconLibrary.GetPixmap();
	if (get_row_height() < pm.height() + 2) {
		set_row_height(pm.height() + 2);
	}

	column(0).set_title("");
	column(1).set_title("Source");
	column_titles_show();
	column_titles_passive();
}

//------------------------------------------------------------------------------

SourceTable::~SourceTable()
{

}

//------------------------------------------------------------------------------

void
SourceTable::AddRow(SourceType nIcon, const std::string& strTitle)
{
	xb::Pixmap* pIcon = NULL;
	
	std::string strFilename;
	if (nIcon == LIBRARY) {
		pIcon = &m_iconLibrary;
	}
	else if (nIcon == CD) {
		pIcon = &m_iconCD;
	}
	else if (nIcon == PLAYLIST) {
		pIcon = &m_iconPlayList;
	}
	else {
		EM::I()->Error() << "invalid icon id: " << nIcon << endl;
		return;
	}

	AssertNotNull(pIcon);
	
	AddRow(pIcon->GetPixmap(), pIcon->GetMask(), strTitle);
	
	SourceInfo info;
	info.m_strName = strTitle;
	info.m_nType = nIcon;
	m_vecSourceInfo.push_back(info);
}

//------------------------------------------------------------------------------

void
SourceTable::UpdateSources(bool bInCallback)
{
	if (!bInCallback) {
		xb::AutoLockGtk lock("UpdateSources");
	}
	
	freeze();
	clear();

	AddRow(SourceTable::LIBRARY, "Library");
	if (System::GetController()->CDAvailable()) {
		std::string strCD("Audio CD");
		if (m_strCDArtist.length() > 0) {
			strCD = m_strCDArtist;
			if (m_strCDTitle.length() > 0) {
				strCD += " - " + m_strCDTitle;
			}
		}
		AddRow(SourceTable::CD, strCD);
	}

	PlayListManager::PlayListVec vec =
		System::GetPlayListManager()->GetPlayListsVec();
	PlayListManager::PlayListVec::const_iterator itr;
	for (itr = vec.begin(); itr != vec.end(); ++itr) {
		EM::I()->Trace() << "Playlist:  " << itr->GetName() << endl;
		AddRow(SourceTable::PLAYLIST, itr->GetName());
	}

	row(0).select();
	
	thaw();

}

//------------------------------------------------------------------------------

void
SourceTable::AddRow(const Gdk_Pixmap& pm, const Gdk_Bitmap& bm,
					const std::string& strTitle)
{
	// xb::AutoLock lock(&s_AddRowLock, "AddRow");

	std::list<std::string> listColumns;
	listColumns.push_back(" ");
	listColumns.push_back(strTitle);

	rows().push_back(listColumns);
	cell(m_nLastRow, 0).set_pixmap(pm, bm);

	m_nLastRow++;
}

//------------------------------------------------------------------------------

std::string
SourceTable::GetName(unsigned int nRow) const
{
	Assert(nRow < m_vecSourceInfo.size(),
		   "SourceTable::GetName called with a row number that is too large");
	
	return m_vecSourceInfo[nRow].m_strName;
}

//------------------------------------------------------------------------------

SourceTable::SourceType
SourceTable::GetType(unsigned int nRow) const
{
	Assert(nRow < m_vecSourceInfo.size(),
		   "SourceTable::GetName called with a row number that is too large");

	return m_vecSourceInfo[nRow].m_nType;
}

//------------------------------------------------------------------------------

void
SourceTable::clear()
{
	m_vecSourceInfo.clear();
	m_nLastRow = 0;
	Gtk::CList::clear();
}

//------------------------------------------------------------------------------

void
SourceTable::DNDReceive(GdkDragContext* pContext, gint x, gint y,
						GtkSelectionData* pData,
						guint info,
						guint32 time)
{
	EM::I()->Noise() << "DNDReceive info = " << info << " length = " <<
		pData->length << " format = " << pData->format << " x = " <<
		x << " y = " << y << endl;

	int nRow, nCol;
	get_selection_info(x, y, &nRow, &nCol);
	EM::I()->Noise() << "row = " << nRow << ", col = " << nCol <<
		", source = " << GetName(nRow) << endl;

	if (GetType(nRow) == PLAYLIST) {

		PlayList* pPlayList = System::GetPlayListManager()->GetPlayList(GetName(nRow));
		if (pPlayList == NULL) {
			EM::I()->Error() << "couldn't get playlist for " << GetName(nRow) <<
				endl;
			return;
		}
		pPlayList->Load();
		
		std::string strData((char*)(pData->data));
		EM::I()->Noise() << "DNDReceive data: " << strData << endl;
		int nPos = 0;
		bool bDone = false;
		while (!bDone) {
			int nNewPos = strData.find_first_of("*", nPos);
			EM::I()->Noise() << "nNewPos " << nNewPos << endl;
			if (nNewPos == -1 || nNewPos == nPos) {
				bDone = true;
			}
			else {
				std::string strFilename = strData.substr(nPos, nNewPos - nPos);
				EM::I()->Noise() << "filename: " << strFilename << endl;

				Song song(strFilename);
				pPlayList->Add(song);
			
				nPos = nNewPos + 1;
			}
		}

		pPlayList->Save();

	}
	else {
		EM::I()->Warn() << "dragged onto something that's not a playlist..." <<
			endl;
	}
}

//------------------------------------------------------------------------------

int
SourceTable::Motion(GdkEventMotion* pMotion)
{
	int nRow, nCol;
	
	get_selection_info(pMotion->x, pMotion->y, &nRow, &nCol);

	Gdk_Color color("#7ca0f9");

	row(nRow).set_background(color);

	return 0;
}

//------------------------------------------------------------------------------

int
SourceTable::DragMotion(GdkDragContext* pDragContext, gint x, gint y,
						guint time)
{
	int nRow, nCol;
	
	get_selection_info(x, y, &nRow, &nCol);

	if (nRow != m_nDragRow) {

		Gdk_Color color("#7ca0f9");
		row(nRow).set_background(color);

		if (m_nDragRow != -1) {
			Gdk_Color white("#ffffff");
			row(m_nDragRow).set_background(white);
		}

		m_nDragRow = nRow;
		
	}
	
	return 0;
}

//------------------------------------------------------------------------------

void
SourceTable::DragLeave(GdkDragContext* pDragContext, guint time)
{
	if (m_nDragRow != -1) {
		Gdk_Color white("#ffffff");
		row(m_nDragRow).set_background(white);
	}

	m_nDragRow = -1;
		
}

//------------------------------------------------------------------------------

void
SourceTable::CDAvailableChange(bool bIn)
{
	UpdateSources(true);
}

//------------------------------------------------------------------------------

void
SourceTable::UpdateCDInfo()
{
	m_strCDArtist = System::GetController()->CDGetArtist();
	m_strCDTitle = System::GetController()->CDGetTitle();
	UpdateSources(true);
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
