// -*-c++-*-
//------------------------------------------------------------------------------
// xtunes - A comprehensive digital music system (http://xtunes.sourceforge.net)
// Copyright 2001,2002 Patrick Crosby <pcrosby@pobox.com>
//------------------------------------------------------------------------------
// Visualization.cpp
//
// $Id: Visualization.cpp,v 1.4 2002/02/18 08:55:59 pcrosby Exp $
//------------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//------------------------------------------------------------------------------

#include "Visualization.h"

#include <glib.h>
#include <xb/ErrorManager.h>

//------------------------------------------------------------------------------

using xb::EM;

#include "Namespace.h"

NAMESPACE_OPEN

//------------------------------------------------------------------------------

Visualization::Visualization()
{

}

//------------------------------------------------------------------------------

Visualization::~Visualization()
{
	ost::MutexLock lock(m_Mutex);

	m_listData.clear();
}

//------------------------------------------------------------------------------

void
Visualization::SetData(int nTime, Output::AFormat nFormat, int nChannels,
					   int nLength, void* pData)
{
	int nMax = nLength / nChannels;
	if (nFormat == Output::FMT_U16_LE || nFormat == Output::FMT_U16_BE ||
		nFormat == Output::FMT_U16_NE || nFormat == Output::FMT_S16_LE ||
		nFormat == Output::FMT_S16_BE || nFormat == Output::FMT_S16_NE) {
		nMax /= 2;
	}

	nMax = CLAMP(nMax, 0, 512);

	Node node;
	node.nTime = nTime;
	node.nChannels = nChannels;
	node.nLength = nMax;


	gint16 	*ptr16;
	guint16	*ptru16;
	guint8	*ptru8;
	gint	i;
	gint max = nMax;
	gint nch = nChannels;
	gpointer ptr = pData;
	
	switch (nFormat)
	{
		case Output::FMT_U8:
			ptru8 = (guint8 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++)
					node.anData[0][i] = ((*ptru8++) ^ 128) << 8;
			else
				for (i = 0; i < max; i++)
				{
					node.anData[0][i] = ((*ptru8++) ^ 128) << 8;
					node.anData[1][i] = ((*ptru8++) ^ 128) << 8;
				}
			break;
		case Output::FMT_S8:
			ptru8 = (guint8 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++)
					node.anData[0][i] = (*ptru8++) << 8;
			else
				for (i = 0; i < max; i++)
				{
					node.anData[0][i] = (*ptru8++) << 8;
					node.anData[1][i] = (*ptru8++) << 8;
				}
			break;
		case Output::FMT_U16_LE:
			ptru16 = (guint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++, ptru16++)
					node.anData[0][i] = GUINT16_FROM_LE(*ptru16) ^ 32768;
			else
				for (i = 0; i < max; i++)
				{
					node.anData[0][i] = GUINT16_FROM_LE(*ptru16) ^ 32768;
					ptru16++;
					node.anData[1][i] = GUINT16_FROM_LE(*ptru16) ^ 32768;
					ptru16++;
				}
			break;
		case Output::FMT_U16_BE:
			ptru16 = (guint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++, ptru16++)
					node.anData[0][i] = GUINT16_FROM_BE(*ptru16) ^ 32768;
			else
				for (i = 0; i < max; i++)
				{
					node.anData[0][i] = GUINT16_FROM_BE(*ptru16) ^ 32768;
					ptru16++;
					node.anData[1][i] = GUINT16_FROM_BE(*ptru16) ^ 32768;
					ptru16++;
				}
			break;
		case Output::FMT_U16_NE:
			ptru16 = (guint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++)
					node.anData[0][i] = (*ptru16++) ^ 32768;
			else
				for (i = 0; i < max; i++)
				{
					node.anData[0][i] = (*ptru16++) ^ 32768;
					node.anData[1][i] = (*ptru16++) ^ 32768;
				}
			break;
		case Output::FMT_S16_LE:
			ptr16 = (gint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++, ptr16++)
					node.anData[0][i] = GINT16_FROM_LE(*ptr16);
			else
				for (i = 0; i < max; i++)
				{
					node.anData[0][i] = GINT16_FROM_LE(*ptr16);
					ptr16++;
					node.anData[1][i] = GINT16_FROM_LE(*ptr16);
					ptr16++;
				}
			break;
		case Output::FMT_S16_BE:
			ptr16 = (gint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++, ptr16++)
					node.anData[0][i] = GINT16_FROM_BE(*ptr16);
			else
				for (i = 0; i < max; i++)
				{
					node.anData[0][i] = GINT16_FROM_BE(*ptr16);
					ptr16++;
					node.anData[1][i] = GINT16_FROM_BE(*ptr16);
					ptr16++;
				}
			break;
		case Output::FMT_S16_NE:
			ptr16 = (gint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++)
					node.anData[0][i] = (*ptr16++);
			else
				for (i = 0; i < max; i++)
				{
					node.anData[0][i] = (*ptr16++);
					node.anData[1][i] = (*ptr16++);
				}
			break;
	}

	ost::MutexLock lock(m_Mutex);
	m_listData.push_back(node);
}

//------------------------------------------------------------------------------

void
Visualization::FindData(int nTime, gint16* pData, int& nLength, int& nChannels)
{
	ost::MutexLock lock(m_Mutex);

	bool bFound = false;
	
	DataList::iterator itr;
	DataList::iterator itrNext;

	itr = m_listData.begin();
	while (!bFound && itr != m_listData.end()) {

		itrNext = itr + 1;


		if (itrNext != m_listData.end()) {
			if (itrNext->nTime >= nTime && itr->nTime < nTime) {

				m_anData = itr->anData;
				nLength = itr->nLength;
				nChannels = itr->nChannels;
				bFound = true;

			}
		}

		if (bFound) {
			if (itr != m_listData.begin()) {
				itr = m_listData.erase(m_listData.begin(), itr - 1);
			}
			else {
				++itr;
			}
		}
		else {
			++itr;
		}
	}
}

//------------------------------------------------------------------------------

void
Visualization::ConvertData(Output::AFormat fmt, gpointer ptr, gint16* left,
						   gint16* right, gint nch, gint max)
{
	gint16 	*ptr16;
	guint16	*ptru16;
	guint8	*ptru8;
	gint	i;
	
	switch (fmt)
	{
		case Output::FMT_U8:
			ptru8 = (guint8 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++)
					left[i] = ((*ptru8++) ^ 128) << 8;
			else
				for (i = 0; i < max; i++)
				{
					left[i] = ((*ptru8++) ^ 128) << 8;
					right[i] = ((*ptru8++) ^ 128) << 8;
				}
			break;
		case Output::FMT_S8:
			ptru8 = (guint8 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++)
					left[i] = (*ptru8++) << 8;
			else
				for (i = 0; i < max; i++)
				{
					left[i] = (*ptru8++) << 8;
					right[i] = (*ptru8++) << 8;
				}
			break;
		case Output::FMT_U16_LE:
			ptru16 = (guint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++, ptru16++)
					left[i] = GUINT16_FROM_LE(*ptru16) ^ 32768;
			else
				for (i = 0; i < max; i++)
				{
					left[i] = GUINT16_FROM_LE(*ptru16) ^ 32768;
					ptru16++;
					right[i] = GUINT16_FROM_LE(*ptru16) ^ 32768;
					ptru16++;
				}
			break;
		case Output::FMT_U16_BE:
			ptru16 = (guint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++, ptru16++)
					left[i] = GUINT16_FROM_BE(*ptru16) ^ 32768;
			else
				for (i = 0; i < max; i++)
				{
					left[i] = GUINT16_FROM_BE(*ptru16) ^ 32768;
					ptru16++;
					right[i] = GUINT16_FROM_BE(*ptru16) ^ 32768;
					ptru16++;
				}
			break;
		case Output::FMT_U16_NE:
			ptru16 = (guint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++)
					left[i] = (*ptru16++) ^ 32768;
			else
				for (i = 0; i < max; i++)
				{
					left[i] = (*ptru16++) ^ 32768;
					right[i] = (*ptru16++) ^ 32768;
				}
			break;
		case Output::FMT_S16_LE:
			ptr16 = (gint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++, ptr16++)
					left[i] = GINT16_FROM_LE(*ptr16);
			else
				for (i = 0; i < max; i++)
				{
					left[i] = GINT16_FROM_LE(*ptr16);
					ptr16++;
					right[i] = GINT16_FROM_LE(*ptr16);
					ptr16++;
				}
			break;
		case Output::FMT_S16_BE:
			ptr16 = (gint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++, ptr16++)
					left[i] = GINT16_FROM_BE(*ptr16);
			else
				for (i = 0; i < max; i++)
				{
					left[i] = GINT16_FROM_BE(*ptr16);
					ptr16++;
					right[i] = GINT16_FROM_BE(*ptr16);
					ptr16++;
				}
			break;
		case Output::FMT_S16_NE:
			ptr16 = (gint16 *)ptr;
			if (nch == 1)
				for (i = 0; i < max; i++)
					left[i] = (*ptr16++);
			else
				for (i = 0; i < max; i++)
				{
					left[i] = (*ptr16++);
					right[i] = (*ptr16++);
				}
			break;
	}
}

//------------------------------------------------------------------------------

const gint16*
Visualization::GetData() const
{
	return (gint16*)m_anData;
}

//------------------------------------------------------------------------------

void
Visualization::Update(int nTime)
{
	int nLength;

	FindData(nTime, (gint16*)m_anData, nLength, m_nChannels);
}

//------------------------------------------------------------------------------

void
Visualization::Reset()
{
	m_listData.clear();
}

//------------------------------------------------------------------------------

NAMESPACE_CLOSE

//------------------------------------------------------------------------------
